
using System;
using Dashboard.ImmediateDraw;
using OpenTK.Mathematics;

namespace Dashboard.Controls
{
    /// <summary>
    /// Bases for all UI elements.
    /// </summary>
    public abstract class UIBase
    {
        private Vector2 size;

        public UIBase? Parent { get; protected set; }
        public string? Id { get; set; }
        public Rectangle Bounds
        {
            get => new Rectangle(Position + Size, Position);
            set
            {
                Size = value.Size;
                Position = value.Min;
            }
        }

        public Vector2 Position { get; set; }

        public Vector2 Size
        {
            get => size;
            set
            {
                Vector2 oldSize = size;
                size = value;

                OnResized(this, new ResizedEventArgs(size, oldSize));
            }
        }

        public Rectangle AbsoluteBounds
        {
            get
            {
                if (Parent == null)
                {
                    return Bounds;
                }
                else
                {
                    return new Rectangle(Bounds.Max + Parent.Position, Bounds.Min + Parent.Position);
                }
            }
        }

        public Vector2 MaximumSize { get; set; } = new Vector2(-1, -1);
        public Vector2 MinimumSize { get; set; } = new Vector2(-1, -1);

        public bool IsMaximumSizeSet => MaximumSize != new Vector2(-1, -1);
        public bool IsMinimumSizeSet => MinimumSize != new Vector2(-1, -1);

        public virtual void NotifyEvent(object? sender, EventArgs args)
        {
        }

        protected virtual void PaintBegin(DrawList cmd)
        {
            cmd.PushViewport();
            cmd.StoreViewport(AbsoluteBounds);
            cmd.PushZ();
        }

        protected virtual void PaintEnd(DrawList cmd)
        {
            cmd.PopViewport();
        }

        public void Paint(DrawList cmd)
        {
            PaintBegin(cmd);
            PaintEnd(cmd);
        }

        public event EventHandler<ResizedEventArgs>? Resized;

        public virtual void OnResized(object sender, ResizedEventArgs ea)
        {
            Resized?.Invoke(sender, ea);
        }
    }

    public class ResizedEventArgs : EventArgs
    {
        public Vector2 NewSize { get; }
        public Vector2 OldSize { get; }

        public ResizedEventArgs(Vector2 newSize, Vector2 oldSize)
        {
            NewSize = newSize;
            OldSize = oldSize;
        }
    }
}